import argparse, os, sys

import pyperclip

from .vm import VM
from .parser_ import Parser, Instruction
from .building import Building, BuildingType

parser = argparse.ArgumentParser(description="Mindustry logic emulator", prog="mlog++/emulator")

parser.add_argument("file", type=str, help="input file [@clip for clipboard]")

args = parser.parse_args()

fn = args.file
code = ""
if fn == "@clip":
    code = pyperclip.paste()
else:
    if not os.path.isfile(fn):
        print(f"Error: input file \"{fn}\" does not exist")
        sys.exit(1)
    
    with open(fn, "r") as f:
        code = f.read()

vm = VM(Parser.parse(code))




import sys
import os

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(os.path.dirname(SCRIPT_DIR), "mlog++"))

from linker import Linker

def test_linker():
    code = Linker.link(["""
wait 0.1""",
"""
print 10
printflush message1
jump 0 always _ _"""])

    assert code == """\
wait 0.1
print 10
printflush message1
jump 1 always _ _"""



import shlex, math, random, re

from .building import Building, BuildingType

REGEX_NUM = re.compile(r"^[0-9]+(\.[0-9]+)?$")

INSTRUCTIONS = {
    "read": 3,
    "write": 3,
    "draw": 7,
    "print": 1,
    "drawflush": 1,
    "printflush": 1,
    "getlink": 2,
    "control": 6,
    "radar": 7,
    "sensor": 3,
    "set": 2,
    "op": 4,
    "end": 0,
    "jump": 4,
    "ubind": 1,
    "ucontrol": 6,
    "uradar": 7,
    "ulocate": 8
}

UNSUPPORTED_INSTRUCTIONS = [
    "draw", "drawflush", "getlink", "control", "radar", "sensor", "ubind", "ucontrol", "uradar", "ulocate"
]

OPERATORS = {
    "add": lambda a, b: a + b,
    "sub": lambda a, b: a - b,
    "mul": lambda a, b: a * b,
    "div": lambda a, b: a / b,
    "idiv": lambda a, b: a // b,
    "mod": lambda a, b: a % b,
    "pow": lambda a, b: a ** b,
    "equal": lambda a, b: a == b,
    "notEqual": lambda a, b: a != b,
    "land": lambda a, b: a and b,
    "lessThan": lambda a, b: a < b,
    "lessThanEq": lambda a, b: a <= b,
    "greaterThan": lambda a, b: a > b,
    "greaterThanEq": lambda a, b: a >= b,
    "strictEqual": lambda a, b: a == b,
    "shl": lambda a, b: a << b,
    "shr": lambda a, b: a >> b,
    "or": lambda a, b: a | b,
    "and": lambda a, b: a & b,
    "xor": lambda a, b: a ^ b,
    "not": lambda a, _: ~a,
    "max": lambda a, b: max(a, b),
    "min": lambda a, b: min(a, b),
    "abs": lambda a, _: abs(a),
    "log": lambda a, _: math.log(a),
    "log10": lambda a, _: math.log10(a),
    "floor": lambda a, _: math.floor(a),
    "ceil": lambda a, _: math.ceil(a),
    "sqrt": lambda a, _: math.sqrt(a),
    "angle": lambda a, b: math.atan2(b, a) * 180 / math.pi,
    "length": lambda a, b: math.sqrt(a * a + b * b),
    "noise": lambda a, b: (random.randint(0, 2000) - 1000) / 1000,
    "rand": lambda a, _: random.randint(0, a),
    "sin": lambda a, _: math.sin(a),
    "cos": lambda a, _: math.cos(a),
    "tan": lambda a, _: math.tan(a),
    "asin": lambda a, _: math.asin(a),
    "acos": lambda a, _: math.acos(a),
    "atan": lambda a, _: math.atan(a)
}

JUMP_CONDITIONS = {
    "equal": lambda a, b: a == b,
    "notEqual": lambda a, b: a != b,
    "lessThan": lambda a, b: a < b,
    "lessThanEq": lambda a, b: a <= b,
    "greaterThan": lambda a, b: a > b,
    "greaterThanEq": lambda a, b: a >= b,
    "strictEqual": lambda a, b: a == b,
    "always": lambda a, b: True
}

class ParserException(Exception):
    pass

class ExecutionError(Exception):
    pass

class Instruction:
    def __init__(self, name: str, params: list = None):
        self.name = name
        self.params = params if params is not None else []
    
    def __repr__(self) -> str:
        params_list = [f"\"{p}\"" for p in self.params]
        return f"EInstruction(\"{self.name}\", [{', '.join(params_list)}])"
    
    def resolve_value(self, value: any, env: dict) -> any:
        if REGEX_NUM.fullmatch(value):
            return float(value)
        
        if type(value) == str:
            if value.startswith("\""):
                if value.endswith("\""):
                    return value[1:-1]
                else:
                    raise ExecutionError(f"Invalid string {value}")
            else:
                if value in env["variables"]:
                    return env["variables"][value]
                else:
                    raise ExecutionError(f"Undefined variable \"{value}\"")

        return value
    
    def execute(self, env: dict):
        if self.name == "read":
            target = self.params[0]
            cell = self.resolve_value(self.params[1], env)
            pos = self.resolve_value(self.params[2], env)

            if type(cell) != Building or cell.type != BuildingType.CELL:
                raise ExecutionError(f"Invalid memory cell \"{cell}\"")
            
            try:
                pos = int(pos)
            except ValueError:
                raise ExecutionError(f"Invalid memory cell position \"{pos}\"")
            
            if pos >= cell["params"]["size"]:
                raise ExecutionError(f"Memory cell access out of bounds {pos}")

            env["variables"][target] = cell["state"]["memory"][pos]
        elif self.name == "write":
            value = self.resolve_value(self.params[0], env)
            cell = self.resolve_value(self.params[1], env)
            pos = self.resolve_value(self.params[2], env)

            try:
                value = float(value)
            except ValueError:
                raise ExecutionError(f"Invalid data to write into memory cell \"{pos}\"")

            if type(cell) != Building or cell.type != BuildingType.CELL:
                raise ExecutionError(f"Invalid memory cell \"{cell}\"")
            
            try:
                pos = int(pos)
            except ValueError:
                raise ExecutionError(f"Invalid memory cell position \"{pos}\"")
            
            if pos >= cell["params"]["size"]:
                raise ExecutionError(f"Memory cell access out of bounds {pos}")
            
            cell["state"]["memory"][pos] = value
        elif self.name == "draw":
            pass
        elif self.name == "print":
            value = self.resolve_value(self.params[0], env)

            env["print_buffer"] += str(value)
        elif self.name == "drawflush":
            pass
        elif self.name == "printflush":
            message = self.resolve_value(self.params[0], env)

            if type(message) != Building or message.type != BuildingType.MESSAGE:
                raise ExecutionError(f"Invalid message \"{message}\"")
            
            message.state["text"] = env["print_buffer"]
            env["print_buffer"] = ""
        elif self.name == "getlink":
            pass
        elif self.name == "control":
            pass
        elif self.name == "radar":
            pass
        elif self.name == "sensor":
            pass
        elif self.name == "set":
            var = self.params[0]
            value = self.resolve_value(self.params[1], env)

            env["variables"][var] = value
        elif self.name == "op":
            op = self.params[0]
            out = self.params[1]
            a = self.resolve_value(self.params[2], env)
            b = self.resolve_value(self.params[3], env)

            if op not in OPERATORS:
                raise ExecutionError(f"Invalid operator \"{op}\"")
            
            try:
                if a not in  [None, True, False]:
                    a = float(a)
            except ValueError:
                raise ExecutionError(f"Invalid operation input [a] \"{a}\"")
            
            try:
                if b not in  [None, True, False]:
                    b = float(b)
            except ValueError:
                raise ExecutionError(f"Invalid operation input [b] \"{b}\"")
            
            env["variables"][out] = OPERATORS[op](a, b)
        elif self.name == "end":
            env["variables"]["@counter"] = 0
        elif self.name == "jump":
            pos = self.params[0]
            try:
                pos = int(pos)
            except ValueError:
                raise ExecutionError(f"Invalid jump destination \"{pos}\"")
            
            cond = self.params[1]
            a = self.resolve_value(self.params[2], env)
            b = self.resolve_value(self.params[2], env)

            if cond not in JUMP_CONDITIONS:
                raise ExecutionError(f"Invalid jump condition \"{cond}\"")

            try:
                if a not in  [None, True, False]:
                    a = float(a)
            except ValueError:
                raise ExecutionError(f"Invalid jump input [a] \"{a}\"")
            
            try:
                if b not in  [None, True, False]:
                    b = float(b)
            except ValueError:
                raise ExecutionError(f"Invalid jump input [b] \"{b}\"")
            
            if JUMP_CONDITIONS[cond](a, b):
                env["variables"]["@counter"] = pos
        elif self.name == "ubind":
            pass
        elif self.name == "ucontrol":
            pass
        elif self.name == "uradar":
            pass
        elif self.name == "ulocate":
            pass

class Parser:
    @staticmethod
    def parse(code: str) -> list:
        tmp = []

        for ln in code.splitlines():
            if not ln.strip():
                continue
            
            spl = shlex.split(ln, comments = False, posix = False)

            if len(spl) > 0:
                if spl[0] in INSTRUCTIONS:
                    if INSTRUCTIONS[spl[0]] != len(spl) - 1:
                        raise ParserException(f"Incorrect number of parameters to instruction ({len(spl) - 1}, expected {INSTRUCTIONS[spl[0]]})")
                    
                    if spl[0] in UNSUPPORTED_INSTRUCTIONS:
                        print(f"WARNING: you are using an unsupported instruction ({spl[0]})")
                    
                    tmp.append(Instruction(spl[0], spl[1:] if len(spl) > 1 else []))
                else:
                    raise ParserException(f"Invalid instruction \"{spl[0]}\"")
        
        print(tmp)
        return tmp



import re, math

from .parser_ import *
from .gerror import gen_error, gen_undefined_error
from . import functions, gerror

GEN_REGEXES = {
    "LABEL": re.compile(r"^<[a-zA-Z_@][a-zA-Z_0-9]*$"),
    "JUMP":  re.compile(r"^>[a-zA-Z_@][a-zA-Z_0-9]*$"),
    "CJMP":  re.compile(r"^>[a-zA-Z_@][a-zA-Z_0-9]* \!?[a-zA-Z_0-9@]+$"),
    "EJMP":  re.compile(r"^>[a-zA-Z_@][a-zA-Z_0-9]* [a-zA-Z_0-9@\.]+ ((==)|(\!=)|(>=)|(<=)|>|<) [a-zA-Z_0-9@\.]+$"),

    "TMPS":  re.compile(r"^set __tmp[0-9]+ .+$"),
    "TMPSE": re.compile(r"^sensor __tmp[0-9]+ \S+ \S+$"),
    "TMPSA": re.compile(r"^set [a-zA-Z_@][a-zA-Z_0-9]* __tmp[0-9]+$"),

    "TMPOP": re.compile(r"^op [a-zA-Z]+( [a-zA-Z_@][a-zA-Z_0-9]*){2} .+$"),

    "MJUMP": re.compile(r"^jump [0-9]+ [a-zA-Z]+ [a-zA-Z_0-9@]+ [a-zA-Z_0-9@]+$"),

    "MCALC": re.compile(r"^op [a-zA-Z]+ [a-zA-Z_0-9@]+ [0-9]+(\.[0-9]+)? [0-9]+(\.[0-9]+)?$"),

    "VARA":  re.compile(r"^(set \S+ \S+)|(op [a-zA-Z]+( \S+){3})$"),

    "NUM":   re.compile(r"^[0-9]+(\.[0-9]+)?$"),

    "IATTR": re.compile(r"^[a-zA-Z_@][a-zA-Z_0-9]*\.[a-zA-Z_@][a-zA-Z_0-9]*$"),

    "TCOMP": re.compile(r"^op ((equal)|(notEqual)|(greaterThan)|(lessThan)|(greaterThanEq)|(lessThanEq)) __tmp[0-9]+ \S+ \S+$"),
    "TJUMP": re.compile(r"^>__mpp[0-9]+ !__tmp[0-9]+$"),

    "VARU":  re.compile(r"^[a-zA-Z_][a-zA-Z_0-9]*$")
}

PRECALC = {
    "add": lambda a, b: a + b,
    "sub": lambda a, b: a - b,
    "mul": lambda a, b: a * b,
    "div": lambda a, b: a / b,
    "idiv": lambda a, b: a // b,
    "mod": lambda a, b: a % b,
    "pow": lambda a, b: a ** b,
    "land": lambda a, b: a and b,
    "lessThan": lambda a, b: a < b,
    "lessThanEq": lambda a, b: a <= b,
    "greaterThan": lambda a, b: a > b,
    "greaterThanEq": lambda a, b: a >= b,
    "strictEqual": lambda a, b: a == b,
    "shl": lambda a, b: a << b,
    "shr": lambda a, b: a >> b,
    "or": lambda a, b: a | b,
    "and": lambda a, b: a & b,
    "xor": lambda a, b: a ^ b,
    "not": lambda a, _: ~a,
    "max": lambda a, b: max(a, b),
    "min": lambda a, b: min(a, b),
    "abs": lambda a, _: abs(a),
    "log": lambda a, _: math.log(a),
    "log10": lambda a, _: math.log10(a),
    "floor": lambda a, _: math.floor(a),
    "ceil": lambda a, _: math.ceil(a),
    "sqrt": lambda a, _: math.sqrt(a)

    # not implemented: angle, length, noise, rand, sin, cos, tan, asin, acos, atan
    # equal and notEqual are not implemented because they use type conversion
}

NATIVE_RETURN_POS = {
    "set": 0,
    "op": 1,
    "read": 0,
    "getlink": 0,
    "sensor": 0,
    "lookup": 1
}

NATIVE_SUB_RETURN_POS = {
    "ucontrol.getBlock": 2,
    "ucontrol.within": 3
}

NATIVE_CONST_INPUTS = {
    "draw": [0],
    "control": [0],
    "sensor": [1],
    "op": [0],
    "jump": [1],
    "radar": [0, 1, 2, 3],
    "ucontrol": [0],
    "uradar": [0, 1, 2, 3],
    "ulocate": [0, 1]
}

PARAM_NO_TMP = {
    "set": [0],
    "op": [1],
    "lookup": [0, 1],
    "read": [0],
    "drawflush": [0],
    "printflush": [0],
    "getlink": [0],
    "radar": [6],
    "sensor": [0],
    "uradar": [6],
    "ubind": [0],
    "ulocate": [4, 5, 6, 7]
}

JUMP_CONDITIONS_REPLACE = {
    "equal": "!=",
    "notEqual": "==",
    "greaterThan": "<=",
    "lessThan": ">=",
    "greaterThanEq": "<",
    "lessThanEq": ">"
}

JUMP_CONDITION = {
    "==": "equal",
    "!=": "notEqual",
    ">": "greaterThan",
    "<": "lessThan",
    ">=": "greaterThanEq",
    "<=": "lessThanEq"
}

class Generator:
    def generate(self, node: AST, optimize_options: dict = None) -> str:
        self.tmpv = 0
        self.tmpl = 0
        self.func_stack = []
        self.loop_stack = []
        self.no_generate_tmp = False
        self.var_list = set(self._generate_var_list(node) + ["true", "false", "null"])

        funcs = []
        for v in self.var_list:
            if ":" in v:
                spl = v.split(":", 1)
                if spl[0] in funcs:
                    gen_error(None, f"Redefinition of function \"{spl[0]}\"")
                funcs.append(spl[0])

        code = self._code_node_join(node)

        return self.postprocess(self.optimize(code, optimize_options))
    
    def _code_join(self, codel: list, to_pos: int = None) -> str:
        code = ""
        for c in codel[:(to_pos + 1 if to_pos is not None else len(codel))]:
            r = self._generate(c)
            code += (r if type(r) == str else r[0]) + "\n"
        
        return "\n".join([l for l in code.strip().splitlines() if l.strip()])
    
    def _code_node_join(self, node: Node, to_pos: int = None) -> str:
        return self._code_join(node.code, to_pos)
    
    def optimize(self, code: str, optimize_options: dict = None) -> str:
        options = {
            "enable": True,
            "unused": True
        }

        if optimize_options is not None:
            for k, v in optimize_options.items():
                options[k] = v
        
        if not options["enable"]:
            return code

        tmp = code

        for _ in range(10):
            for _ in range(1, 101):
                tmp, found = self._single_tmp_optimize(tmp, 1)
                if not found:
                    break
            
            for i in range(1, 11):
                tmp, found = self._single_tmp_optimize(tmp, i)
            
            for _ in range(1, 11):
                tmp, found = self._negative_optimize(tmp)

        for i in range(1, 101):
            tmp, found = self._forward_use_optimize(tmp, i)
            if not found:
                break
        
        for _ in range(10):
            tmp, found = self._precalc_optimize(tmp)

            for i in range(1, 11):
                tmp, found = self._single_tmp_optimize(tmp, i)
            
            for i in range(1, 101):
                tmp, found = self._forward_use_optimize(tmp, i)
                if not found:
                    break
        
        if options["unused"]:
            for _ in range(1, 101):
                tmp, found = self._unused_optimize(tmp)
                if not found:
                    break

        return tmp
    
    def _single_tmp_optimize(self, code: str, n_lines: int) -> str:
        uses = {}

        for i in range(1, self.tmpv + 1):
            c = len(re.findall(f"__tmp{i}\\D", code))

            if c > 0:
                uses[f"__tmp{i}"] = c
        
        found = False
        lns = code.splitlines()
        tmp = ""
        for i, ln in enumerate(lns):
            if GEN_REGEXES["TMPS"].fullmatch(ln):
                spl = ln.split(" ", 2)
                name = spl[1]
                val = spl[2]

                if i < len(lns) - n_lines:
                    if uses[name] == 2 and name in lns[i + n_lines]:
                            lns[i + n_lines] = lns[i + n_lines].replace(name, val)
                            found = True
                            continue

            elif GEN_REGEXES["TMPSE"].fullmatch(ln):
                spl = ln.split(" ", 3)
                name = spl[1]

                if i < len(lns) - n_lines:
                    if uses[name] == 2 and GEN_REGEXES["TMPSA"].fullmatch(lns[i + n_lines]):
                        spl_ = lns[i + n_lines].split(" ", 2)
                        if name == spl_[2]:
                            lns[i + n_lines] = f"sensor {spl_[1]} {spl[2]} {spl[3]}"
                            found = True
                            continue
            
            elif GEN_REGEXES["TCOMP"].fullmatch(ln):
                spl = ln.split(" ", 4)
                cond = spl[1]
                name = spl[2]
                op1 = spl[3]
                op2 = spl[4]

                if i < len(lns) - n_lines:
                    if uses[name] == 2 and GEN_REGEXES["TJUMP"].fullmatch(lns[i + n_lines]):
                        spl_ = lns[i + n_lines].split(" ", 1)
                        if spl_[1][1:] == spl[2] and cond in JUMP_CONDITIONS_REPLACE:
                            lns[i + n_lines] = f"{spl_[0]} {op1} {JUMP_CONDITIONS_REPLACE[cond]} {op2}"
                            continue
                
            tmp += ln + "\n"
        
        return tmp, found
    
    def _forward_use_optimize(self, code: str, n_lines: int) -> str:
        lns = code.splitlines()
        tmp = ""
        found = False
        for i, ln in enumerate(lns):
            if GEN_REGEXES["TMPS"].fullmatch(ln):
                spl = ln.split(" ", 2)
                name = spl[1]
                val = spl[2]

                if i < len(lns) - n_lines:
                    if GEN_REGEXES["TMPOP"].fullmatch(lns[i + n_lines]):
                        spl2 = lns[i + n_lines].split(" ", 4)
                        if spl2[2] == name:
                            lns[i + n_lines] = f"{spl2[0]} {spl2[1]} {spl2[2]} {val} {spl2[4]}"
                            found = True
                            continue
            tmp += ln + "\n"
        
        return tmp, found
    
    def _precalc_optimize(self, code: str) -> str:
        lns = code.splitlines()
        tmp = ""
        found = False
        for ln in lns:
            if GEN_REGEXES["MCALC"].fullmatch(ln):
                spl = ln.split(" ", 4)
                op = spl[1]
                name = spl[2]
                n1 = float(spl[3])
                n2 = float(spl[4])

                if op in PRECALC:
                    result = PRECALC[op](n1, n2)
                    if result == int(result):
                        result = int(result)
                    
                    ln = f"set {name} {result}"
                    found = True

            tmp += ln + "\n"

        return tmp, found
    
    def _unused_optimize(self, code: str) -> str:
        tmp = ""
        found = False
        for ln in code.splitlines():
            if GEN_REGEXES["VARA"].fullmatch(ln):
                name = ""

                if ln.startswith("set"):
                    name = ln.split(" ", 2)[1]
                elif ln.startswith("op"):
                    name = ln.split(" ", 4)[2]
                
                uses = len(re.findall(name, code))

                if uses <= 1:
                    continue

            tmp += ln + "\n"

        return tmp, found
    
    def _negative_optimize(self, code: str) -> str:
        lns = code.splitlines()
        tmp = ""
        found = False
        for i, ln in enumerate(lns):
            if GEN_REGEXES["TMPS"].fullmatch(ln):
                if i < len(lns) - 1:
                    spl = ln.split(" ", 2)
                    name = spl[1]
                    val = spl[2]

                    if lns[i + 1] == f"op sub {name} 0 {name}":
                        if GEN_REGEXES["NUM"].fullmatch(val):
                            lns[i + 1] = f"set {name} -{val}"
                            continue

            tmp += ln + "\n"
        
        return tmp, found
    
    def postprocess(self, code: str) -> str:
        if not code.strip():
            return code

        tmp = ""
        labels = {}
        lc = 0

        for i, ln in enumerate(code.splitlines()):
            if ln.startswith("."):
                ln = ln[1:]
            
            if GEN_REGEXES["LABEL"].fullmatch(ln):
                labels[ln[1:]] = i - lc
                lc += 1
                continue

            if GEN_REGEXES["VARA"]:
                spl = ln.split()
                if spl and spl[0] == "set":
                    if spl[1] == spl[2]:
                        continue
                
            
            tmp += ln + "\n"
        
        tmp2 = ""
        for i, ln in enumerate(tmp.splitlines()):
            name = ""
            cond = False
            cvar = "_"
            invert = False

            if GEN_REGEXES["JUMP"].fullmatch(ln):
                name = ln[1:]
            elif GEN_REGEXES["CJMP"].fullmatch(ln):
                spl = ln.split()
                name = spl[0][1:]
                cond = True
                invert = spl[1].startswith("!")
                cvar = spl[1][1:] if invert else spl[1]
            elif GEN_REGEXES["EJMP"].fullmatch(ln):
                spl = ln.split()
                name = spl[0][1:]
                op1 = spl[1]
                cond = spl[2]

                if not cond in JUMP_CONDITION:
                    gen_error(None, f"Unknown jump condition \"{cond}\"")
                cond = JUMP_CONDITION[cond]

                op2 = spl[3]
                if name in labels:
                    tmp2 += f"jump {labels[name]} {cond} {op1} {op2}\n"
                else:
                    gen_error(None, f"Unknown label \"{name}\"")
                continue
            
            if name == "":
                tmp2 += ln + "\n"
            else:
                if name in labels:
                    tmp2 += f"jump {labels[name]} {'always' if not cond else 'notEqual' if invert else 'equal'} {cvar} true\n"
                else:
                    gen_error(None, f"Unknown label \"{name}\"")
        
        tmp = ""
        lnc = len(tmp2.strip().splitlines())
        for i, ln in enumerate(tmp2.splitlines()):
            if GEN_REGEXES["MJUMP"].fullmatch(ln):
                spl = ln.split(" ", 4)
                if int(spl[1]) >= lnc:
                    ln = f"{spl[0]} 0 {spl[2]} {spl[3]} {spl[4]}"

            tmp += ln + "\n"
        
        if tmp.splitlines()[-1].startswith("jump 0"):
            tmp = "\n".join(tmp.splitlines()[:-1])
        
        return tmp.strip()
    
    def _generate(self, node: Node):
        t = type(node)

        if t == Node:
            gen_error(node, "Invalid node")
        elif t == CodeNode:
            return "\n".join([str(self._generate(c)) for c in node.code])
        elif t == ValueNode:
            if self.no_generate_tmp:
                var = str(node.value)
            else:
                var = self.get_tmp_var()

            if GEN_REGEXES["IATTR"].fullmatch(node.value):
                spl = node.value.split(".")
                
                return f"sensor {var} {spl[0]} @{spl[1]}", var
            
            if type(node.value) == str and not (node.value.startswith("\"") and node.value.endswith("\"")) and node.value != "_" and node.value[-1] not in "0123456789":
                if GEN_REGEXES["VARU"].fullmatch(node.value) and node.value not in self.var_list and node.value not in functions.builtin:
                    gen_undefined_error(node, node.value)

            return f"set {var} {node.value}" if not self.no_generate_tmp else "", var
        elif t == IndexNode:
            var = self.get_tmp_var()

            tmp, var_ = self._generate(node.index)
            
            return f"{tmp}\nread {var} {node.var} {var_}", var
        elif t == AtomNode:
            return self._generate(node.value)
        elif t == AssignmentNode:
            if node.atype == "=":
                tmp, var = self._generate(node.right)

                if GEN_REGEXES["IATTR"].fullmatch(node.left):
                    spl = node.left.split(".", 1)
                    return f"{tmp}\ncontrol {spl[1]} {spl[0]} {var} _ _ _"

                return f"{tmp}\nset {node.left} {var}"
            elif node.atype in ["+=", "-=", "*=", "/="]:
                at = node.atype
                op = "add" if at == "+=" else "sub" if at == "-=" else "mul" if at == "*=" else "div" if at == "/=" else ""
                if op == "":
                    gen_error(node, f"Invalid operator: \"{at}\"")
                
                tmp, var = self._generate(node.right)
                return f"{tmp}\nop {op} {node.left} {node.left} {var}"
        elif t == IndexAssignNode:
            if node.atype == "=":
                itmp, ivar = self._generate(node.index)
                vtmp, vvar = self._generate(node.val)

                return f"{itmp}\n{vtmp}\nwrite {vvar} {node.var} {ivar}"
            elif node.atype in ["+=", "-=", "*=", "/="]:
                itmp, ivar = self._generate(node.index)
                vtmp, vvar = self._generate(node.val)

                at = node.atype
                op = "add" if at == "+=" else "sub" if at == "-=" else "mul" if at == "*=" else "div" if at == "/=" else ""
                if op == "":
                    gen_error(node, f"Invalid operator: \"{at}\"")
                
                tval = self.get_tmp_var()

                return f"{itmp}\n{vtmp}\nread {tval} {node.var} {ivar}\nop {op} {tval} {tval} {vvar}\nwrite {tval} {node.var} {ivar}"
        elif t == ExpressionNode:
            tmp, var = self._generate(node.left)

            if node.right is not None:
                for r in node.right:
                    op = "land" if r[0] == "&&" else "or" if r[0] == "||" else ""
                    if op == "":
                        gen_error(node, f"Invalid operator: \"{r[0]}\"")
                    
                    tmp2, var2 = self._generate(r[1])
                    tmp += f"\n{tmp2}\nop {op} {var} {var} {var2}"

            return tmp, var
        elif t == CompExpressionNode:
            tmp, var = self._generate(node.left)

            if node.right is not None:
                for r in node.right:
                    op = "equal" if r[0] == "==" else "lessThan" if r[0] == "<" else "greaterThan" if r[0] == ">" else "lessThanEq" if r[0] == "<=" \
                        else "greaterThanEq" if r[0] == ">=" else "notEqual" if r[0] == "!=" else "strictEqual" if r[0] == "===" else ""
                    if op == "":
                        gen_error(node, f"Invalid operator: \"{r[0]}\"")
                    
                    tmp2, var2 = self._generate(r[1])
                    tmp += f"\n{tmp2}\nop {op} {var} {var} {var2}"

            return tmp, var
        elif t == ArithExpNode:
            tmp, var = self._generate(node.left)

            if node.right is not None:
                for r in node.right:
                    op = "add" if r[0] == "+" else "sub" if r[0] == "-" else ""
                    if op == "":
                        gen_error(node, f"Invalid operator: \"{r[0]}\"")
                    
                    tmp2, var2 = self._generate(r[1])
                    tmp += f"\n{tmp2}\nop {op} {var} {var} {var2}"

            return tmp, var
        elif t == TermNode:
            tmp, var = self._generate(node.left)

            if node.right is not None:
                for r in node.right:
                    op = "mul" if r[0] == "*" else "div" if r[0] == "/" else "pow" if r[0] == "**" else ""
                    if op == "":
                        gen_error(node, f"Invalid operator: \"{r[0]}\"")
                    
                    tmp2, var2 = self._generate(r[1])
                    tmp += f"\n{tmp2}\nop {op} {var} {var} {var2}"

            return tmp, var
        elif t == FactorNode:
            tmp, var = self._generate(node.left)

            if not node.sign:
                tmp += f"\nop sub {var} 0 {var}"
            if node.not_:
                tmp += f"\nop not {var} {var} _"
            
            return tmp, var
        elif t == CallNode:
            if node.is_call:
                if type(node.function) == AtomNode:
                    node.function = str(node.function.value.value)
                
                if type(node.function) != str:
                        node.function = self._generate(node.function)[0].split()[-1]

                is_native = node.function in functions.native
                is_builtin = node.function in functions.builtin

                if not is_native and not is_builtin:
                    signature = functions.gen_signature(node.function, node.params)
                    found = False

                    for v in self.var_list:
                        if v == signature:
                            found = True
                            break

                        if ":" in v:
                            spl = v.split(":")
                            if spl[0] == node.function:
                                gen_error(node, f"Incorrect number of parameters to function (expected {spl[1]}, got {len(node.params)})")
                    
                    if not found:
                        gen_error(node, f"Undefined function \"{node.function}\"")

                tmp = ""
                params = []
                for i, arg in enumerate(node.params):
                    pnt = PARAM_NO_TMP.get(node.function, [])
                    
                    if i in pnt:
                        onogen = self.no_generate_tmp
                        self.no_generate_tmp = True
                    
                    nci = NATIVE_CONST_INPUTS.get(node.function, [])

                    if i in nci:
                        gerror.push_undefined(False)
                    
                    tmp2, var = self._generate(arg)
                    params.append(var)

                    if i in pnt:
                        self.no_generate_tmp = onogen
                    
                    if i in nci:
                        gerror.pop_undefined()

                    if is_native or is_builtin:
                        tmp += f"{tmp2}\n"
                    else:
                        tmp += f"{tmp2}\nset __f_{node.function}_arg_{i} {var}\n"
                
                if is_native:
                    req = functions.native_params[node.function]
                    while len(params) < req:
                        params.append("_")
                    
                    retpos = NATIVE_RETURN_POS.get(node.function, -1)
                    retvar = "null"
                    if retpos != -1:
                        retvar = params[retpos]
                    
                    return f"{tmp}{node.function} {' '.join(params)}", retvar
                elif is_builtin:
                    nparams = functions.builtin_params.get(node.function, functions.builtin_params_default)
                    if nparams != len(params):
                        gen_error(node, f"Incorrect number of parameters to function (expected {nparams}, got {len(params)})")

                    return f"{tmp}op {node.function} __f_{node.function}_retv {params[0] if nparams >= 1 else '_'} {params[1] if nparams >= 2 else '_'}", f"__f_{node.function}_retv"
                else:
                    return f"{tmp}op add __f_{node.function}_ret @counter 1\nset @counter __f_{node.function}", f"__f_{node.function}_retv"
            else:
                tmp, var = self._generate(node.function)

                return tmp, var
        elif t == SubCallNode:
            spl = node.function.split(".")
            req = functions.native_params[spl[0]]

            tmp = ""
            params = []
            for i, arg in enumerate(node.params):
                tmp_, var = self._generate(arg)

                tmp += f"{tmp_}\n"
                params.append(var)
            
            while len(params) < req - 1:
                params.append("_")
            
            retpos = NATIVE_SUB_RETURN_POS.get(node.function, -1)
            retvar = "null"
            if retpos != -1:
                retvar = params[retpos]
            
            return f"{tmp}{spl[0]} {spl[1]} {' '.join(params)}", retvar
        elif t == IfNode:
            tmp2, var = self._generate(node.condition)

            if not node.elsecode:
                l_e = self.get_tmp_label()

                tmp = f"{tmp2}\n>{l_e} !{var}\n"
                tmp += self._code_node_join(node)
                tmp += f"\n<{l_e}"
            else:
                l_s = self.get_tmp_label()
                l_e = self.get_tmp_label()

                tmp = f"{tmp2}\n>{l_s} !{var}\n"
                tmp += self._code_node_join(node)
                tmp += f"\n>{l_e}\n<{l_s}\n"
                tmp += self._code_join(node.elsecode)
                tmp += f"\n<{l_e}"

            return tmp
        elif t == WhileNode:
            tmp2, var = self._generate(node.condition)

            l_v = self.get_tmp_label()
            l_e = self.get_tmp_label()

            self.loop_stack.append((l_v, l_e, Node()))

            tmp = f"<{l_v}\n{tmp2}\n>{l_e} !{var}\n"
            tmp += self._code_node_join(node)
            tmp += f"\n>{l_v}\n<{l_e}"

            self.loop_stack.pop()

            return tmp
        elif t == ForNode:
            itmp = self._generate(node.init)
            ctmp, cvar = self._generate(node.condition)
            atmp = self._generate(node.action)

            if type(itmp) != str:
                itmp = itmp[0]
            
            if type(atmp) != str:
                atmp = atmp[0]

            l_v = self.get_tmp_label()
            l_e = self.get_tmp_label()

            self.loop_stack.append((l_v, l_e, node.action))

            tmp = f"{itmp}\n<{l_v}\n{ctmp}\n>{l_e} !{cvar}\n"
            tmp += self._code_node_join(node)
            tmp += f"\n{atmp}\n>{l_v}\n<{l_e}"

            self.loop_stack.pop()

            return tmp
        elif t == FunctionNode:
            l_e = self.get_tmp_label()

            args = {a: f"__f_{node.name}_arg_{i}" for i, a in enumerate(node.args)}

            self.func_stack.append(node.name)

            to_pos = None
            for i, n in enumerate(node.code):
                if type(n) == ReturnNode:
                    to_pos = i
                    break

            tmp = f"op add __f_{node.name} @counter 1\n>{l_e}\n"
            tmp += self._code_node_join(node.rrename(args), to_pos)

            if to_pos is None:
                tmp += f"\nset @counter __f_{node.name}_ret"
            
            tmp += f"\n<{l_e}"

            self.func_stack.pop()

            return tmp
        elif t == RepeatNode:
            if node.amount == 0:
                return ""

            l_v = self.get_tmp_label()
            l_e = self.get_tmp_label()
            i = self.get_tmp_var()
            gen = self._code_node_join(node)

            tmp = f"op add {i} 0 1\n<{l_v}\nop add {i} {i} 1\n"
            tmp += gen
            tmp += f"\n>{l_e} {i} == {node.amount}\n>{l_v}\n<{l_e}"

            if len(self.optimize(tmp).strip().splitlines()) - 2 >= node.amount * len(self.optimize(gen).strip().splitlines()):
                tmp = gen
                for _ in range(node.amount - 1):
                    tmp += "\n" + gen

            return tmp
        elif t == NativeNode:
            return f".{node.code}"
        elif t == ReturnNode:
            if len(self.func_stack) < 1:
                gen_error(node, "Cannot return when not in a function")
            
            fname = self.func_stack[-1]
            rvar = f"__f_{fname}_retv"
            if node.value == None:
                return f"set {rvar} null", rvar
            else:
                tmp, var = self._generate(node.value)
                return f"{tmp}\nset {rvar} {var}\nset @counter __f_{fname}_ret", rvar
        elif t == LoopActionNode:
            if len(self.loop_stack) < 1:
                gen_error(node, "Cannot break or continue when not in a loop")
            
            loop = self.loop_stack[-1]

            if node.action == "break":
                return f">{loop[1]}"
            elif node.action == "continue":
                return f"{self._generate(loop[2])}\n>{loop[0]}"
            
            raise RuntimeError(f"Invalid AST")
        elif t == ExternNode:
            return ""
        
        gen_error(node, f"Unknown node ({node})")
    
    def _var_list_join(self, lists: list) -> list:
        tmp = []
        for l in lists:
            tmp += l
        return tmp
    
    def _generate_var_list(self, node: Node) -> list:
        t = type(node)
        
        if t in [AST, CodeNode]:
            return self._var_list_join([self._generate_var_list(n) for n in node.code])
        elif t == ValueNode:
            return [node.value] if type(node.value) == str and not (node.value.startswith("\"") and node.value.endswith("\"")) else []
        elif t == IndexNode:
            return self._generate_var_list(node.index)
        elif t == AtomNode:
            return self._generate_var_list(node.value)
        elif t == AssignmentNode:
            return [node.left] if node.atype == "=" else [] + self._generate_var_list(node.right)
        elif t == IndexAssignNode:
            return []
        elif t in [ExpressionNode, CompExpressionNode, ArithExpNode, TermNode]:
            tmp = []
            tmp += self._generate_var_list(node.left)
            if node.right is not None:
                for _, e in node.right:
                    tmp += self._generate_var_list(e)
            return tmp
        elif t == FactorNode:
            return self._generate_var_list(node.left)
        elif t == CallNode:
            if node.is_call:
                tmp = []
                for i in PARAM_NO_TMP.get(node.function, []):
                    if i < len(node.params):
                        vl = self._generate_var_list(node.params[i])
                        tmp += vl
                return [f"__f_{node.function}_retv"] + tmp
            else:
                return self._generate_var_list(node.function)
        elif t == SubCallNode:
            if node.function in NATIVE_SUB_RETURN_POS:
                return self._generate_var_list(node.params[NATIVE_SUB_RETURN_POS[node.function]])
            
            return []
        elif t in [IfNode, WhileNode, ForNode, RepeatNode, FunctionNode]:
            tmp = self._var_list_join([self._generate_var_list(n) for n in node.code])
            if t in [IfNode, WhileNode, ForNode]:
                tmp += self._generate_var_list(node.condition)
            if t == IfNode and node.elsecode is not None:
                tmp += self._var_list_join([self._generate_var_list(n) for n in node.elsecode])
            if t == ForNode:
                tmp += self._generate_var_list(node.init)
                tmp += self._generate_var_list(node.action)
            if t == FunctionNode:
                tmp += [f"__f_{node.name}_arg_{i}" for i, _ in enumerate(node.args)]
                tmp += [functions.gen_signature(node.name, node.args)]
            return tmp
        elif t == NativeNode:
            return []
        elif t == ReturnNode:
            return self._generate_var_list(node.value)
        elif t == LoopActionNode:
            return []
        elif t == ExternNode:
            return [node.name]

        gen_error(node, f"Unknown node ({node})")

    def get_tmp_var(self) -> str:
        self.tmpv += 1
        return f"__tmp{self.tmpv}"
    
    def get_tmp_label(self) -> str:
        self.tmpl += 1
        return f"__mpp{self.tmpl}"



from enum import Enum

class BuildingType(Enum):
    CELL      = "cell"
    MESSAGE   = "message"
    PROCESSOR = "processor"

class Building:
    def __init__(self, type_: BuildingType, name: str, params: dict):
        self.type = type_
        self.name = name
        self.params = params
        self.state = {}

        if type_ == BuildingType.CELL:
            self.state["memory"] = [0 for _ in range(params["size"])]
        elif type_ == BuildingType.MESSAGE:
            self.state["text"] = ""
        elif type_ == BuildingType.PROCESSOR:
            self.state[""]



def gen_signature(name: str, params: list):
    return f"{name}:{len(params)}"

native = [
    "read", "write",
    "draw", "drawflush",
    "print", "printflush",
    "getlink",
    "control",
    "radar",
    "sensor",
    "set", "op",
    "wait", "lookup",
    "end", "jump",
    "ubind", "ucontrol", "uradar", "ulocate"
]

native_params = {
    "read": 3, "write": 3,
    "draw": 7, "drawflush": 1,
    "print": 1, "printflush": 1,
    "getlink": 2,
    "control": 6,
    "radar": 6,
    "sensor": 3,
    "set": 2, "op": 4,
    "wait": 1, "lookup": 3,
    "end": 0, "jump": 4,
    "ubind": 1, "ucontrol":6, "uradar": 6, "ulocate": 8
}

native_sub = {
    "draw": {
        "clear": 3,
        "color": 4,
        "stroke": 1,
        "line": 4,
        "rect": 4,
        "lineRect": 4,
        "poly": 5,
        "linePoly": 5,
        "triangle": 6,
        "image": 5
    },
    "control": {
        "enabled": 2,
        "shoot": 4,
        "shootp": 3,
        "config": 2,
        "color": 4
    },
    "ucontrol": {
        "idle": 0,
        "stop": 0,
        "move": 2,
        "approach": 3,
        "boost": 1,
        "pathfind": 0,
        "target": 3,
        "targetp": 2,
        "itemDrop": 2,
        "itemTake": 3,
        "payDrop": 0,
        "payTake": 1,
        "payEnter": 0,
        "mine": 2,
        "flag": 1,
        "build": 5,
        "getBlock": 4,
        "within": 4
    }
}

builtin = [
    "mod",
    "pow",
    "and", "or", "xor", "not",
    "max", "min",
    "abs",
    "log", "log10",
    "ceil", "floor",
    "sqrt",
    "sin", "cos", "tan",
    "asin", "acos", "atan",
    "len"
]

builtin_params_default = 1
builtin_params = {
    "mod": 2,
    "pow": 2,
    "and": 2,
    "or": 2,
    "xor": 2,
    "max": 2,
    "min": 2,
    "len": 2
}

keywords = ["if", "else", "while", "for", "function", "repeat"]

special = native + builtin + keywords



import sys, inspect

from .lexer import Token, Position
from . import arrows
from .formatting import Format
from .parser_ import Node

GEN_ERROR_DEBUG  = False
GEN_U_ERROR_DEBUG = True
GEN_F_ERROR_DEBUG = False

_undefined_stack = []

def push_undefined(value: bool) -> None:
    _undefined_stack.append(value)

def pop_undefined() -> bool:
    return _undefined_stack.pop()

def gen_error(node: Node, msg: str) -> None:
    if GEN_ERROR_DEBUG:
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        print(f"In method: {calframe[1][3]}, line: {calframe[1][2]}\n")

    if node is not None:
        print(f"{Format.ERROR}Generator error on line {node.pos.line}, column {node.pos.column}: {msg}{Format.RESET}\n")
        print(f"Here:\n{Token._sanitize(node.pos.cline)}\n{arrows.generate(node.pos.column, node.pos.len)}")
    else:
        print(f"{Format.ERROR}Generator error: {msg}{Format.RESET}")

    sys.exit(1)

def gen_undefined_error(node: Node, var: str) -> None:
    if len(_undefined_stack) > 0 and not _undefined_stack[-1]:
        return
    
    if GEN_U_ERROR_DEBUG:
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        print(f"In method: {calframe[1][3]}, line: {calframe[1][2]}\n")
    
    if node is not None:
        print(f"{Format.ERROR}Generator error on line {node.pos.line}, column {node.pos.column}: Undefined variable {var}{Format.RESET}\n")
        print(f"Here:\n{Token._sanitize(node.pos.cline)}\n{arrows.generate(node.pos.column, node.pos.len)}")
    else:
        print(f"{Format.ERROR}Generator error: Undefined variable {var}{Format.RESET}")
    
    sys.exit(1)

def gen_undefinedf_error(node: Node, name: str) -> None:
    if GEN_F_ERROR_DEBUG:
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        print(f"In method: {calframe[1][3]}, line: {calframe[1][2]}\n")
    
    if node is not None:
        print(f"{Format.ERROR}Generator error on line {node.pos.line}, column {node.pos.column}: Undefined function {name}{Format.RESET}\n")
        print(f"Here:\n{Token._sanitize(node.pos.cline)}\n{arrows.generate(node.pos.column, node.pos.len)}")
    else:
        print(f"{Format.ERROR}Generator error: Undefined function {name}{Format.RESET}")
    
    sys.exit(1)



from .parser_ import Instruction

class VMSignal(Exception):
    pass

class VM:
    def __init__(self, ins: list):
        self.ins = ins
        self.env = {
            "variables": {
                "_": 0,
                "null": None,
                "true": True,
                "false": False,

                # Special variables
                "@this": None,
                "@thisx": 0,
                "@thisy": 0,
                "@ipt": 0,
                "@counter": 0,
                "@links": 0,
                "@unit": None,
                "@time": 0,
                "@tick": 0,
                "@mapw": 0,
                "@maph": 0
            },

            "print_buffer": ""
        }
    
    def next_ins(self) -> Instruction:
        if self.env["variables"]["@counter"] < len(self.ins) - 1:
            return self.ins[self.env["variables"]["@counter"]]
        
        raise VMSignal()
    
    def step(self) -> bool:
        try:
            i = self.next_ins()
        except VMSignal:
            return False
        
        self.env["variables"]["@counter"] += 1
        i.execute(self.env)
        
        return True
    
    def cycle(self) -> dict:
        while self.step():
            pass

        return self.env



import sys
import os

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(os.path.dirname(SCRIPT_DIR), "mlog++"))

from lexer import Token, TokenType, Lexer

def test_lexer():
    t = Lexer.lex("""\
const a = test
function a(
# parameters
@_1x,__2x){
print(1**2+3-4==1)
}""")
    assert t[4].type == TokenType.ID and t[4].value == "function"
    assert t[7].type == TokenType.ID and t[7].value == "@_1x"

    t = Lexer.preprocess(t)

    print(Lexer.stringify_tokens(t))

    assert t[1].type == TokenType.ID and t[1].value == "test"



from .lexer import TokenType, Token, Position
from .error import parse_error
from . import functions

class Node:
    def generate(self, i = 0) -> str:
        return i * " " + "Node\n"

class AST(Node):
    def __init__(self, code: list):
        self.code = code
    
    def generate(self, i = 0) -> str:
        tmp = i * " " + "AST {\n"
        for c in self.code:
            tmp += c.generate(i + 1)
        tmp += i * " " + "}\n"
        return tmp
    
    def rrename(self, vars_: dict):
        return AST([n.rrename(vars_) for n in self.code])

class CodeNode(Node):
    def __init__(self, pos: Position, code: list):
        self.pos = pos
        self.code = code
    
    def generate(self, i) -> str:
        tmp = i * " " + "CodeNode {\n"
        for c in self.code:
            tmp += c.generate(i + 1)
        tmp += i * " " + "}\n"
        return tmp
    
    def rrename(self, vars_: dict) -> Node:
        return CodeNode(self.pos, [n.rrename(vars_) for n in self.code])

class ValueNode(Node):
    def __init__(self, pos: Position, value: str):
        self.pos = pos
        self.value = value
    
    def generate(self, i) -> str:
        return i * " " + "ValueNode: " + str(self.value) + "\n"
    
    def rrename(self, vars_: dict) -> Node:
        if "." in self.value:
            spl = self.value.split(".")
            if len(spl) == 2:
                return ValueNode(self.pos, vars_.get(spl[0], spl[0]) + "." + spl[1])
        
        return ValueNode(self.pos, vars_.get(self.value, self.value))

class IndexNode(Node):
    def __init__(self, pos: Position, var: str, index: Node):
        self.pos = pos
        self.var = var
        self.index = index
    
    def generate(self, i) -> str:
        return i * " " + "IndexNode: " + str(self.var) + ", " + self.index.generate(i + 1) + "\n"
    
    def rrename(self, vars_: dict) -> Node:
        return IndexNode(self.pos, vars_.get(self.var, self.var), self.index.rrename(vars_))

class IndexAssignNode(Node):
    def __init__(self, pos: Position, var: str, index: Node, atype: str, val: Node):
        self.pos = pos
        self.var = var
        self.index = index
        self.atype = atype
        self.val = val
    
    def generate(self, i) -> str:
        return i * " " + "IndexNode: " + str(self.var) + ", " + self.index.generate(i + 1) + ", " + self.atype + ", " + self.val.generate(i + 1) + "\n"
    
    def rrename(self, vars_: dict) -> Node:
        return IndexNode(self.pos, vars_.get(self.var, self.var), self.index.rrename(vars_), self.atype, self.val.rrename(vars_))

class AtomNode(Node):
    def __init__(self, pos: Position, value):
        self.pos = pos
        self.value = value
    
    def generate(self, i) -> str:
        return i * " " + "AtomNode: \n" + self.value.generate(i + 1)
    
    def rrename(self, vars_: dict) -> Node:
        return self.value.rrename(vars_)

class AssignmentNode(Node):
    def __init__(self, pos: Position, left: str, atype: str, right: Node):
        self.pos = pos
        self.left = left
        self.atype = atype
        self.right = right
    
    def generate(self, i) -> str:
        return i * " " + "AssignmentNode: \n" + (i + 1) * " " + f"{self.left} {self.atype} {self.right.generate(0)}"
    
    def rrename(self, vars_: dict) -> Node:
        return AssignmentNode(self.pos, vars_.get(self.left, self.left), self.atype, self.right.rrename(vars_))

class ExpressionNode(Node):
    def __init__(self, pos: Position, left, right = None):
        self.pos = pos
        self.left = left
        self.right = right
    
    def generate(self, i) -> str:
        tmp = ""
        if self.right is not None:
            for r in self.right:
                tmp += f" {r[0]} {r[1].generate(i + 1)}"
        return i * " " + f"ExpressionNode: \n{self.left.generate(i + 1)}{tmp}"
    
    def rrename(self, vars_: dict) -> Node:
        return ExpressionNode(self.pos, self.left.rrename(vars_), [[n[0], n[1].rrename(vars_)] for n in self.right] if self.right is not None else None)

class CompExpressionNode(Node):
    def __init__(self, pos: Position, left, right = None):
        self.pos = pos
        self.left = left
        self.right = right
    
    def generate(self, i) -> str:
        tmp = ""
        if self.right is not None:
            for r in self.right:
                tmp += f" {r[0]} {r[1].generate(i + 1)}"
        return i * " " + f"CompExpressionNode: \n{self.left.generate(i + 1)}{tmp}"
    
    def rrename(self, vars_: dict) -> Node:
        return CompExpressionNode(self.pos, self.left.rrename(vars_), [[n[0], n[1].rrename(vars_)] for n in self.right] if self.right is not None else None)

class ArithExpNode(Node):
    def __init__(self, pos: Position, left, right = None):
        self.pos = pos
        self.left = left
        self.right = right
    
    def generate(self, i) -> str:
        tmp = ""
        if self.right is not None:
            for r in self.right:
                tmp += f" {r[0]} {r[1].generate(i + 1)}"
        return i * " " + f"ArithExpNode: \n{self.left.generate(i + 1)}{tmp}"
    
    def rrename(self, vars_: dict) -> Node:
        return ArithExpNode(self.pos, self.left.rrename(vars_), [[n[0], n[1].rrename(vars_)] for n in self.right] if self.right is not None else None)

class TermNode(Node):
    def __init__(self, pos: Position, left, right = None):
        self.pos = pos
        self.left = left
        self.right = right
    
    def generate(self, i) -> str:
        tmp = ""
        if self.right is not None:
            for r in self.right:
                tmp += f" {r[0]} {r[1].generate(i + 1)}"
        return i * " " + f"TermNode: \n{self.left.generate(i + 1)}{tmp}"
    
    def rrename(self, vars_: dict) -> Node:
        return TermNode(self.pos, self.left.rrename(vars_), [[n[0], n[1].rrename(vars_)] for n in self.right] if self.right is not None else None)

class FactorNode(Node):
    def __init__(self, pos: Position, left, sign: bool, not_: bool):
        self.pos = pos
        self.left = left
        self.sign = sign
        self.not_ = not_
    
    def generate(self, i) -> str:
        return i * " " + f"FactorNode: \n{'!' if self.not_ else ''}{'+' if self.sign else '-'}{self.left.generate(i + 1)}"
    
    def rrename(self, vars_: dict) -> Node:
        return FactorNode(self.pos, self.left.rrename(vars_), self.sign, self.not_)

class CallNode(Node):
    def __init__(self, pos: Position, function, is_call: bool, params: list = []):
        self.pos = pos
        self.function = function
        self.is_call = is_call
        self.params = params
    
    def generate(self, i) -> str:
        if self.is_call:
            return i * " " + f"CallNode [1]: {self.function}({','.join(self.params)})\n"
        else:
            return i * " " + f"CallNode [0]: {self.function.generate(i + 1)}\n"
    
    def rrename(self, vars_: dict) -> Node:
        return CallNode(self.pos, self.function if type(self.function) == str else self.function.rrename(vars_), self.is_call, [n.rrename(vars_) for n in self.params])

class SubCallNode(Node):
    def __init__(self, pos: Position, function: str, params: list = []):
        self.pos = pos
        self.function = function
        self.params = params
    
    def generate(self, i) -> str:
        return i * " " + f"SubCallNode: {self.function}({','.join(self.params)})\n"
    
    def rrename(self, vars_: dict) -> Node:
        return SubCallNode(self.pos, self.function, [n.rrename(vars_) for n in self.params])


class KeywordNode(Node):
    def __init__(self):
        pass

class IfNode(KeywordNode):
    def __init__(self, pos: Position, condition: ExpressionNode, code: list, elsecode: list = None):
        self.pos = pos
        self.condition = condition
        self.code = code
        self.elsecode = elsecode if elsecode is not None else []
    
    def generate(self, i) -> str:
        tmp = i * " " + f"IfNode ({self.condition.generate(i + 1)}) {{"
        for c in self.code:
            tmp += c.generate(i + 1)
        tmp += i * " " + "} {\n"
        for c in self.elsecode:
            tmp += c.generate(i + 1)
        tmp += i * " " + "}\n"
        return tmp
    
    def rrename(self, vars_: dict) -> Node:
        return IfNode(self.pos, self.condition.rrename(vars_), [n.rrename(vars_) for n in self.code], [n.rrename(vars_) for n in self.elsecode])

class WhileNode(KeywordNode):
    def __init__(self, pos: Position, condition: ExpressionNode, code: list):
        self.pos = pos
        self.condition = condition
        self.code = code
    
    def generate(self, i) -> str:
        tmp = i * " " + f"WhileNode ({self.condition.generate(i + 1)}) {{"
        for c in self.code:
            tmp += c.generate(i + 1)
        tmp += i * " " + "}\n"
        return tmp
    
    def rrename(self, vars_: dict) -> Node:
        return WhileNode(self.pos, self.condition.rrename(vars_), [n.rrename(vars_) for n in self.code])

class ForNode(KeywordNode):
    def __init__(self, pos: Position, init: Node, condition: ExpressionNode, action: Node, code: list):
        self.pos = pos
        self.init = init
        self.condition = condition
        self.action = action
        self.code = code
    
    def generate(self, i) -> str:
        tmp = i * " " + f"ForNode ({self.init.generate(i + 1)}; {self.condition.generate(i + 1)}; {self.action.generate(i + 1)}) {{"
        for c in self.code:
            tmp += c.generate(i + 1)
        tmp += i * " " + "}\n"
        return tmp
    
    def rrename(self, vars_: dict) -> Node:
        return ForNode(self.pos, self.init.rrename(vars_), self.condition.rrename(vars_), self.action.rrename(vars_), [n.rrename(vars_) for n in self.code])

class FunctionNode(KeywordNode):
    def __init__(self, pos: Position, name: str, args: list, code: list):
        self.pos = pos
        self.name = name
        self.args = args
        self.code = code
    
    def rrename(self, vars_: dict) -> Node:
        return FunctionNode(self.pos, self.name, self.args, [n.rrename(vars_) for n in self.code])

class RepeatNode(KeywordNode):
    def __init__(self, pos: Position, amount: int, code: list):
        self.pos = pos
        self.amount = amount
        self.code = code
    
    def generate(self, i) -> str:
        tmp = i * " " + f"RepeatNode ({self.amount}) {{"
        for c in self.code:
            tmp += c.generate(i + 1)
        tmp += i * " " + "}\n"
        return tmp
    
    def rrename(self, vars_: dict) -> Node:
        return RepeatNode(self.pos, self.amount, [n.rrename(vars_) for n in self.code])

class NativeNode(KeywordNode):
    def __init__(self, pos: Position, code: str):
        self.pos = pos
        self.code = code
    
    def rrename(self, vars_: dict) -> Node:
        return NativeNode(self.pos, self.code)

class ReturnNode(KeywordNode):
    def __init__(self, pos: Position, value):
        self.pos = pos
        self.value = value
    
    def rrename(self, vars_: dict) -> Node:
        return ReturnNode(self.pos, self.value.rrename(vars_))

class LoopActionNode(KeywordNode):
    def __init__(self, pos: Position, action: str):
        self.pos = pos
        self.action = action
    
    def rrename(self, vars_: dict) -> Node:
        return LoopActionNode(self.pos, self.action)

class ExternNode(Node):
    def __init__(self, pos: Position, name: str):
        self.pos = pos
        self.name = name
    
    def rrename(self, vars_: dict) -> Node:
        return ExternNode(self.pos, self.name)

class Parser:
    def parse(self, tokens: list) -> CodeNode:
        self.tokens = tokens
        self.pos = -1

        nodes = []
        while self.has_token():
            nodes.append(self.parse_AnyNode())

        return AST(nodes)
    
    def has_token(self) -> bool:
        return self.pos < len(self.tokens) - 1
    
    def curr_token(self) -> Token:
        return self.tokens[self.pos]
    
    def next_token(self, ttype=None) -> Token:
        if self.has_token():
            self.pos += 1

            if self.tokens[self.pos].type != ttype and ttype is not None:
                parse_error(self.tokens[self.pos], "Unexpected token")

            return self.tokens[self.pos]
        
        parse_error(self.tokens[-1], "Unexpected EOF")
    
    # node parsers

    def parse_AnyNode(self, can_be_special=True) -> Node:
        id_ = self.next_token()
        if id_.type == TokenType.ID:
            n = self.next_token()

            if id_.value == "return":
                self.pos -= 1
                return ReturnNode(id_.pos(), self.parse_ExpressionNode())
            elif id_.value in ["break", "continue"]:
                self.pos -= 1
                return LoopActionNode(id_.pos(), id_.value)
            elif id_.value == "extern":
                if n.type != TokenType.ID:
                    parse_error(n, "Unexpected token")
                
                n_ = self.next_token()
                if n_.type == TokenType.LPAREN:
                    args = []
                    last = TokenType.LPAREN
                    n__ = n
                    while True:
                        n = self.next_token()
                        if n.type == TokenType.RPAREN:
                            if last == TokenType.COMMA:
                                parse_error(n, "Unexpected token")
                            break
                        elif n.type == TokenType.COMMA:
                            if last == TokenType.COMMA:
                                parse_error(n, "Unexpected token")
                            pass
                        elif n.type == TokenType.ID:
                            if last == TokenType.ID:
                                parse_error(n, "Unexpected token")
                            args.append(n.value)
                    return ExternNode(id_.pos(), functions.gen_signature(n__.value, args))
                
                self.pos -= 1
                return ExternNode(id_.pos(), n.value)
            
            if n.type == TokenType.SET:
                if id_.value in functions.special:
                    parse_error(n, "Cannot override a keyword")
                
                return AssignmentNode(id_.pos(), id_.value, n.value, self.parse_ExpressionNode())
            elif n.type == TokenType.DOT:
                if id_.value in functions.native_sub:
                    n_ = self.next_token(TokenType.ID)
                    
                    if not n_.value in functions.native_sub[id_.value]:
                        parse_error(n_, "Invalid subcommand")

                    self.next_token(TokenType.LPAREN)

                    e = False
                    p = []
                    while True:
                        t = self.next_token()
                        if t.type == TokenType.RPAREN:
                            break
                        elif t.type == TokenType.COMMA:
                            if e:
                                p.append(self.parse_ExpressionNode())
                            else:
                                parse_error(t, "Unexpected token")
                        else:
                            if not e:
                                self.pos -= 1
                                p.append(self.parse_ExpressionNode())
                                e = True
                            else:
                                parse_error(t, "Unexpected token")
                    
                    return SubCallNode(id_.pos(), f"{id_.value}.{n_.value}", p)

                if id_.value in functions.special:
                    parse_error(n, "Cannot use property of a keyword")

                node = self.parse_AnyNode()
                if type(node) != AssignmentNode:
                    parse_error(n, "Unexpected token")
                
                node.left = id_.value + "." + node.left

                return node
            elif n.type == TokenType.LBRACK:
                if id_.value in functions.special:
                    parse_error(n, "Cannot use keyword as a memory cell")
                
                node = self.parse_ExpressionNode()

                self.next_token(TokenType.RBRACK)
                a = self.next_token(TokenType.SET)

                val = self.parse_ExpressionNode()

                return IndexAssignNode(id_.pos(), id_.value, node, a.value, val)
            elif n.type == TokenType.LPAREN:
                if id_.value in functions.keywords:
                    if not can_be_special:
                        parse_error(id_, "Unexpected token")
                    
                    self.pos -= 2
                    return self.parse_KeywordNode()
                else:
                    e = False
                    p = []
                    while True:
                        t = self.next_token()
                        if t.type == TokenType.RPAREN:
                            break
                        elif t.type == TokenType.COMMA:
                            if e:
                                p.append(self.parse_ExpressionNode())
                            else:
                                parse_error(t, "Unexpected token")
                        else:
                            if not e:
                                self.pos -= 1
                                p.append(self.parse_ExpressionNode())
                                e = True
                            else:
                                parse_error(t, "Unexpected token")
                    
                    return CallNode(id_.pos(), id_.value, True, p)
            elif n.type == TokenType.ID:
                if id_.value == "function":
                    self.pos -= 2
                    return self.parse_KeywordNode()
                else:
                    parse_error(n, "Unexpected token")
            elif n.type == TokenType.LBRACE:
                if id_.value == "else":
                    self.pos -= 2
                    return self.parse_KeywordNode()
                else:
                     parse_error(n, "Unexpected token")
            else:
                parse_error(n, "Unexpected token")
        elif id_.type == TokenType.DOT:
            n = self.next_token(TokenType.STRING)

            return NativeNode(id_.pos(), n.value[1:-1])
        
        parse_error(id_, "Unexpected token")
    
    def parse_ExpressionNode(self) -> ExpressionNode:
        c = self.parse_CompExpressionNode()
        l = []
        while self.has_token():
            n = self.next_token()
            if n.type == TokenType.LOGIC and n.value in ["&&", "||"]:
                l.append((n.value, self.parse_CompExpressionNode()))
            else:
                self.pos -= 1
                break
        
        return ExpressionNode(c.pos, c, l if len(l) > 0 else None)

    def parse_CompExpressionNode(self) -> CompExpressionNode:
        e = self.parse_ArithExpNode()
        c = []
        while self.has_token():
            n = self.next_token()
            if n.type == TokenType.OPERATOR and n.value in ["==", "!=", "<", ">", "<=", ">=", "==="]:
                c.append((n.value, self.parse_ArithExpNode()))
            else:
                self.pos -= 1
                break
        
        return CompExpressionNode(e.pos, e, c if len(c) > 0 else None)
    
    def parse_ArithExpNode(self) -> ArithExpNode:
        t = self.parse_TermNode()
        a = []
        while self.has_token():
            n = self.next_token()
            if n.type == TokenType.OPERATOR and n.value in ["+", "-"]:
                a.append((n.value, self.parse_TermNode()))
            else:
                self.pos -= 1
                break
        
        return ArithExpNode(t.pos, t, a if len(a) > 0 else None)
    
    def parse_TermNode(self) -> TermNode:
        f = self.parse_FactorNode()
        m = []
        while self.has_token():
            n = self.next_token()
            if n.type == TokenType.OPERATOR and n.value in ["*", "/", "**"]:
                m.append((n.value, self.parse_FactorNode()))
            else:
                self.pos -= 1
                break
        
        return TermNode(f.pos, f, m if len(m) > 0 else None)
    
    def parse_FactorNode(self) -> FactorNode:
        sign = True
        not_ = False
        while self.has_token():
            n = self.next_token()
            if n.type == TokenType.OPERATOR and n.value in ["+", "-", "!"]:
                if n.value == "-":
                    sign = not sign
                elif n.value == "!":
                    not_ = not not_
            elif n.type in [TokenType.ID, TokenType.STRING, TokenType.NUMBER, TokenType.LPAREN]:
                self.pos -= 1
                return FactorNode(n.pos(), self.parse_CallNode(), sign, not_)
            else:
                parse_error(n, "Unexpected token")
    
    def parse_CallNode(self) -> CallNode:
        a = self.parse_AtomNode()
        p = []
        is_call = False

        if self.has_token():
            n = self.next_token()
            if n.type == TokenType.LPAREN:
                is_call = True

                if not self.has_token():
                    parse_error(n, "Unexpected EOF")
                
                n = self.next_token()
                if n.type == TokenType.RPAREN:
                    pass
                else:
                    self.pos -= 1
                    while self.has_token():
                        p.append(self.parse_ExpressionNode())
                        n = self.next_token()
                        if n.type == TokenType.RPAREN:
                            break
                        elif n.type == TokenType.COMMA:
                            pass
                        else:
                            parse_error(n, "Unexpected token")
            else:
                self.pos -= 1
        
        return CallNode(a.pos, a, is_call, p)
    
    def parse_AtomNode(self) -> AtomNode:
        n = self.next_token()
        
        if n.type in [TokenType.ID, TokenType.STRING, TokenType.NUMBER]:
            if self.has_token() and n.type == TokenType.ID:
                n_ = self.next_token()
                if n_.type == TokenType.DOT:
                    n_ = self.next_token(TokenType.ID)
                    return AtomNode(n.pos(), ValueNode(n.pos(), f"{n.value}.{n_.value}"))
                elif n_.type == TokenType.LBRACK:
                    n_ = self.parse_ExpressionNode()
                    self.next_token(TokenType.RBRACK)
                    return AtomNode(n.pos(), IndexNode(n.pos(), n.value, n_))
                else:
                    self.pos -= 1
            
            return AtomNode(n.pos(), ValueNode(n.pos(), n.value))
        elif n.type == TokenType.LPAREN:
            e = self.parse_ExpressionNode()
            self.next_token()
            return AtomNode(n.pos(), e)
        else:
            parse_error(n, "Unexpected token")
    
    def parse_KeywordNode(self) -> KeywordNode:
        id_ = self.next_token()

        if id_.value in ["if", "while"]:
            self.next_token(TokenType.LPAREN)
            e = self.parse_ExpressionNode()
            self.next_token(TokenType.RPAREN)

            c = []
            self.next_token()
            while True:
                n = self.next_token()
                if n.type == TokenType.RBRACE:
                    break
                else:
                    self.pos -= 1
                    c.append(self.parse_AnyNode())
            
            if id_.value == "if":
                if self.has_token():
                    n = self.next_token()
                    if n.type == TokenType.ID and n.value == "else":
                        self.next_token(TokenType.LBRACE)

                        ec = []
                        while True:
                            n = self.next_token()
                            if n.type == TokenType.RBRACE:
                                break
                            else:
                                self.pos -= 1
                                ec.append(self.parse_AnyNode())
                        
                        return IfNode(id_.pos(), e, c, ec)
                    else:
                        self.pos -= 1

                return IfNode(id_.pos(), e, c)
            elif id_.value == "while":
                return WhileNode(id_.pos(), e, c)
        elif id_.value == "for":
            self.next_token(TokenType.LPAREN)

            init = self.parse_AnyNode(False)

            self.next_token(TokenType.SEMICOLON)
            cond = self.parse_ExpressionNode()

            self.next_token(TokenType.SEMICOLON)
            action = self.parse_AnyNode(False)

            self.next_token(TokenType.RPAREN)

            c = []
            self.next_token(TokenType.LBRACE)
            while True:
                n = self.next_token()
                if n.type == TokenType.RBRACE:
                    break
                else:
                    self.pos -= 1
                    c.append(self.parse_AnyNode())

            return ForNode(id_.pos(), init, cond, action, c)
        elif id_.value == "function":
            id_ = self.next_token(TokenType.ID)

            self.next_token(TokenType.LPAREN)

            args = []
            last = TokenType.LPAREN
            while True:
                n = self.next_token()
                if n.type == TokenType.RPAREN:
                    if last == TokenType.COMMA:
                        parse_error(n, "Unexpected token")
                    break
                elif n.type == TokenType.COMMA:
                    if last == TokenType.COMMA:
                        parse_error(n, "Unexpected token")
                    pass
                elif n.type == TokenType.ID:
                    if last == TokenType.ID:
                        parse_error(n, "Unexpected token")
                    args.append(n.value)

            c = []
            self.next_token(TokenType.LBRACE)
            while True:
                n = self.next_token()
                if n.type == TokenType.RBRACE:
                    break
                else:
                    self.pos -= 1
                    c.append(self.parse_AnyNode())
            
            return FunctionNode(id_.pos(), id_.value, args, c)
        elif id_.value == "repeat":
            self.next_token(TokenType.LPAREN)
            amount = self.next_token(TokenType.NUMBER)
            self.next_token(TokenType.RPAREN)

            c = []
            self.next_token(TokenType.LBRACE)
            while True:
                n = self.next_token()
                if n.type == TokenType.RBRACE:
                    break
                else:
                    self.pos -= 1
                    c.append(self.parse_AnyNode())
            
            return RepeatNode(id_.pos(), int(amount.value), c)
        
        parse_error(id_, "Unexpected token")



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("mlogpp/__init__.py", encoding="utf-8") as f:
    exec(f.read())

setuptools.setup(
    name="mlogpp",
    version=__version__,
    author="albi-c",
    description="mlog++ to mindustry logic compiler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/albi-c/mlog++",
    project_urls={
        "Bug Tracker": "https://github.com/albi-c/mlog++/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(),
    python_requires=">=3.6"
)



from .lexer import Position
from .error import link_error

class Linker:
    def link(codes: list) -> str:
        offset = 0
        tmp = []
        for code in codes:
            c, o = Linker._relocate(code, offset)

            tmp.append(c)
            offset += o

        return "\n".join(tmp)
    
    def _relocate(code: str, offset: int) -> str:
        tmp = ""
        nl = 0
        for ln in code.strip().splitlines():
            if ln.startswith("jump "):
                spl = ln.split(" ")

                if len(spl) != 5:
                    link_error(Position(nl, 0, ln, len(ln)), "Invalid jump instruction")

                pos = spl[1]
                try:
                    pos = int(pos)
                except ValueError:
                    link_error(Position(nl, len(spl[0]) + 1, ln, len(spl[1])), "Invalid jump address")
                
                ln = f"jump {pos + offset} {spl[2]} {spl[3]} {spl[4]}"

            nl += 1
            tmp += ln + "\n"
        
        return tmp.strip(), nl



import argparse, os, sys
from enum import Enum

import pyperclip

from .lexer import Lexer
from .parser_ import Parser
from .generator import Generator
from .linker import Linker
from . import __version__

class IOMethod(Enum):
    FILE = 0
    STD  = 1
    CLIP = 2

parser = argparse.ArgumentParser(description="Mindustry logic compiler", prog="mlog++")

parser.add_argument("file", type=str, help="input file(s) [@clip for clipboard]", nargs="+")

parser.add_argument("-o:f", "--output-file", help="write output to a file")
parser.add_argument("-o:s", "--output-stdout", help="write output to stdout", action="store_true")
parser.add_argument("-o:c", "--output-clip", help="write output to clipboard (defualt)", action="store_true")

parser.add_argument("-O0", "--optimize0", help="disable optimization (WARNING: creates extremely unoptimized code)", action="store_true")
parser.add_argument("-O1", "--optimize1", help="set optimization level to 1", action="store_true")
parser.add_argument("-O2", "--optimize2", help="set optimization level to 2 (default)", action="store_true")

parser.add_argument("-v", "--verbose", help="print additional information", action="store_true")
parser.add_argument("-l", "--lines", help="print line numbers when output to stdout is selected", action="store_true")

parser.add_argument("-V", "--version", action="version", version=f"mlog++ {__version__}")

args = parser.parse_args()

omethod = IOMethod.CLIP
ofile = ""

optlevel = 2

verbose = False

for k, v in vars(args).items():
    if v:
        if k.startswith("output"):
            omethod = IOMethod.FILE if k.endswith("file") else IOMethod.STD if k.endswith("stdout") else IOMethod.CLIP
            if omethod == IOMethod.FILE:
                ofile = v
        elif k.startswith("optimize"):
            optlevel = int(k[-1])
        elif k == "verbose":
            verbose = v

for fn in args.file:
    if not os.path.isfile(fn) and fn != "@clip":
        print(f"Error: input file \"{fn}\" does not exist")
        sys.exit(1)

datas = []
for fn in args.file:
    if fn == "@clip":
        datas.append((fn, pyperclip.paste()))
    else:
        with open(fn, "r") as f:
            datas.append((fn, f.read()))

optimization_levels = {
    0: {"enable": False},
    1: {"unused": False},
    2: {}
}

outs = []
for data in datas:
    if data[0].endswith(".mind"):
        outs.append(data[1])
        continue

    out = Lexer.preprocess(Lexer.lex(Lexer.resolve_includes(data[1])))
    out = Parser().parse(out)
    out = Generator().generate(out, optimization_levels[optlevel])
    outs.append(out)

if len(outs) > 1:
    out = Linker.link(outs).strip()
else:
    out = outs[0].strip()

if omethod == IOMethod.FILE:
    with open(ofile, "w+") as f:
        f.write(out)
elif omethod == IOMethod.STD:
    if vars(args)["lines"]:
        lines = out.splitlines()
        maxline = len(str(len(lines)))
        for i, ln in enumerate(lines):
            print(f"{str(i).zfill(maxline)}: {ln}")
    else:
        print(out)
    
    if verbose:
        print()
elif omethod == IOMethod.CLIP:
    pyperclip.copy(out)

if verbose:
    print(f"Output: {len(out.strip())} characters, {len(out.strip().split())} words, {len(out.strip().splitlines())} lines")



import sys
import os

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(os.path.dirname(SCRIPT_DIR), "mlog++"))

import arrows

def test_arrows():
    assert arrows.generate(4, 1) == "    ^"
    assert arrows.generate(0, 5) == "^^^^^"



# import sys
# import os

# SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
# sys.path.append(os.path.join(os.path.dirname(SCRIPT_DIR), "mlog++"))

from lexer import Lexer
from parser_ import Parser
from generator import Generator

def test_generator():
    tokens = Lexer.preprocess(Lexer.lex("""
    const a = 10
    const b = 20
    print(10)
    print(a + b)
    if (a === b) {
        print(20)
    }
    printflush(message1)
"""))

    ast = Parser().parse(tokens)
    code = Generator().generate(ast)

    assert code == """\
print 10
print 30
jump 4 notEqual 0 true
print 20
printflush message1"""



import sys, inspect

from .lexer import Token, Position
from . import arrows
from .formatting import Format

PARSE_ERROR_DEBUG = False
LINK_ERROR_DEBUG  = False

def parse_error(tok: Token, msg: str) -> None:
    if PARSE_ERROR_DEBUG:
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        print(f"In method: {calframe[1][3]}, line: {calframe[1][2]}\n")

    print(f"{Format.ERROR}Parser error on line {tok.line}, column {tok.col}: {msg}{Format.RESET}\n")
    print(f"Here:\n{tok.sanitize_cline()}\n{arrows.generate(tok.col, len(tok.value))}")

    sys.exit(1)

def link_error(pos: Position, msg: str) -> None:
    if LINK_ERROR_DEBUG:
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        print(f"In method: {calframe[1][3]}, line: {calframe[1][2]}\n")
    
    print(f"{Format.ERROR}Linker error on line {pos.line}, column {pos.column}: {msg}{Format.RESET}\n")
    print(f"Here:\n{Token._sanitize(pos.cline)}\n{arrows.generate(pos.column, pos.len)}")

    sys.exit(1)



from enum import Enum
import re, sys, os

from . import arrows
from .formatting import Format

class TokenType(Enum):
    NONE      = 0
    ID        = 2
    STRING    = 3
    NUMBER    = 4
    LPAREN    = 5
    RPAREN    = 6
    LBRACE    = 7
    RBRACE    = 8
    LBRACK    = 9
    RBRACK    = 10
    COMMA     = 11
    SEMICOLON = 12
    OPERATOR  = 13
    SET       = 14
    LOGIC     = 15
    DOT       = 16

LEX_REGEXES = {
    TokenType.ID: re.compile(r"^[a-zA-Z_@][a-zA-Z_0-9]*$"),
    TokenType.STRING: re.compile("^\"([^\"\\\\]|\\\\.)*\"$"),
    TokenType.NUMBER: re.compile(r"^[0-9]+(\.[0-9]+)?$"),
    TokenType.LPAREN: re.compile(r"^\($"),
    TokenType.RPAREN: re.compile(r"^\)$"),
    TokenType.LBRACE: re.compile(r"^\{$"),
    TokenType.RBRACE: re.compile(r"^\}$"),
    TokenType.LBRACK: re.compile(r"^\[$"),
    TokenType.RBRACK: re.compile(r"^\]$"),
    TokenType.COMMA: re.compile(r"^,$"),
    TokenType.SEMICOLON: re.compile(r"^;$"),
    TokenType.OPERATOR: re.compile(r"^[+\-*/!]|(\*\*)|(===)|(<=)|(>=)|(==)|(\!=)|<|>$"),
    TokenType.SET: re.compile(r"^=|(\+=)|(\-=)|(\*=)|(\/=)$"),
    TokenType.LOGIC: re.compile(r"^(\&\&)|(\|\|)$"),
    TokenType.DOT: re.compile(r"^\.$")
}

class Position:
    def __init__(self, line: int, column: int, cline: str, len_: int):
        self.line = line
        self.column = column
        self.cline = cline
        self.len = len_
    
    def __repr__(self) -> str:
        return f"Position({self.line}, {self.column}, \"{Token._sanitize(self.cline)}\" {self.len})"
    
    def __str__(self) -> str:
        return f"Position: [line: {self.line}, column: {self.column}, code line: \"{Token._sanitize(self.cline)}\", length: {self.len}]"
    
    def __eq__(self, other) -> bool:
        return self.line == other.line and self.column == other.column

class Token:
    def __init__(self, type_: TokenType, value: str, line: int, col: int, cline: str):
        self.type = type_
        self.value = value
        self.line = line
        self.col = col
        self.cline = cline
    
    def _sanitize(s: str) -> str:
        return s.replace("\n", "\\n").replace("\t", "\\t").replace("\r", "\\r")
        
    def sanitize_value(self) -> str:
        return Token._sanitize(self.value)
    
    def sanitize_cline(self) -> str:
        return Token._sanitize(self.cline)
    
    def pos(self) -> Position:
        return Position(self.line, self.col, self.cline, len(self.value))
    
    def __repr__(self) -> str:
        return f"Token({self.type}, \"{self.sanitize_value()}\", {self.line}, {self.col}, \"{self.sanitize_cline()}\")"
    
    def __str__(self) -> str:
        return f"Token: [type: {self.type}, value: \"{self.sanitize_value()}\", line: {self.line}, column: {self.col}, code line: \"{self.sanitize_cline()}\"]"
    
    def __eq__(self, other) -> bool:
        return self.type == other.type and self.value == other.value

class Lexer:
    @staticmethod
    def resolve_includes(code: str) -> str:
        tmp = ""
        for i, ln in enumerate(code.splitlines()):
            if ln.startswith("%"):
                fn = ln[1:]

                if not os.path.isfile(fn):
                    print(f"{Format.ERROR}Error on line {i + 1}: Cannot import file \"{fn}\"{Format.RESET}\n\nHere:\n{ln}\n{arrows.generate(0, len(ln))}")
                    sys.exit(1)
                
                with open(fn, "r") as f:
                    data = f.read()
                
                tmp += data + "\n"
                continue
            
            tmp += ln + "\n"
        
        return tmp

    @staticmethod
    def lex(code: str) -> list:
        toks = []
        tok = ""
        in_str = False

        for li, ln in enumerate(code.splitlines()):
            st = ln.strip()

            if not st or st.startswith("#"):
                continue
            
            if st.startswith(">") or st.startswith("<"):
                ln = f".{st}"

            st = ln.strip()
            if st.startswith("."):
                ln = f".\"{st[1:]}\""
            
            prev = ""
            for i, c in enumerate(ln):
                if c == " " and not in_str:
                    if Lexer._match(tok) == TokenType.NONE and tok.strip():
                        print(f"{Format.ERROR}Error on line {li + 1}, column {i + 1}: Invalid token \"{tok}\"{Format.RESET}\n\nHere:\n{ln}\n{arrows.generate(i - len(tok), len(tok))}")
                        sys.exit(1)
                    elif tok.strip():
                        toks.append(Token(Lexer._match(tok), tok, li + 1, i - len(tok), ln))
                        tok = ""
                    
                    continue
                
                if c == "\"" and prev != "\\":
                    in_str = not in_str
                
                matched = Lexer._match(tok)

                tok += c

                matches = Lexer._match(tok)

                if tok == "==" and i < len(ln) - 1:
                    if ln[i + 1] == "=":
                        toks.append(Token(TokenType.OPERATOR, "===", li, i - len(tok), ln))
                        tok = ""
                        continue
                elif c == "." and matched == TokenType.NUMBER and matches == TokenType.NONE:
                    continue

                if ((matched != TokenType.NONE) and (matches == TokenType.NONE)) or ((matched != matches) and matched != TokenType.NONE):
                    if tok in ["+=", "-=", "*=", "/=", ">=", "<=", "==", "**"]:
                        toks.append(Token(Lexer._match(tok), tok, li + 1, i - len(tok), ln))
                        tok = ""
                    else:
                        tok = tok[:-1]
                        toks.append(Token(Lexer._match(tok), tok, li + 1, i - len(tok), ln))
                        tok = c

                prev = c

                if len(toks) >= 2:
                    if toks[-1].value == "=" and toks[-2].value == "===":
                        toks.pop()
            
            if Lexer._match(tok) != TokenType.NONE:
                toks.append(Token(Lexer._match(tok), tok, li + 1, i - len(tok), ln))
                tok = ""
            elif tok.strip():
                print(f"{Format.ERROR}Error on line {li + 1}, column {i + 1}: Invalid token \"{tok}\"{Format.RESET}\n\nHere:\n{ln}\n{arrows.generate(i - len(tok), len(tok))}")
                sys.exit(1)

        return toks

    @staticmethod
    def _match(token: str) -> TokenType:
        for t, r in LEX_REGEXES.items():
            if r.fullmatch(token):
                return t
        
        return TokenType.NONE
    
    @staticmethod
    def _matchtp(tokens: list, pos: int, pattern: list, *patterns: list) -> bool:
        for p in patterns:
            if len(p) != len(pattern):
                return False

        if pos + len(pattern) > len(tokens):
            return False
        
        matches = [t.type for t in tokens[pos:pos+len(pattern)]] == pattern

        for p in patterns:
            if [t.type for t in tokens[pos:pos+len(p)]] == p:
                matches = True
        
        return matches
    
    @staticmethod
    def stringify_tokens(tokens: list) -> str:
        return "\n".join([str(t) for t in tokens])
    
    @staticmethod
    def preprocess(tokens: list) -> list:
        tmp = tokens.copy()

        consts = {}
        found = True
        while found:
            found = False

            for i, t in enumerate(tmp):
                if t.value == "const" and Lexer._matchtp(tmp, i, [TokenType.ID, TokenType.ID, TokenType.SET, TokenType.NUMBER],
                                                                [TokenType.ID, TokenType.ID, TokenType.SET, TokenType.STRING],
                                                                [TokenType.ID, TokenType.ID, TokenType.SET, TokenType.ID]):
                    
                    consts[tmp[i + 1].value] = tmp[i + 3]

                    for _ in range(4):
                        tmp.pop(i)

                    found = True

                    break
        
        for i, t in enumerate(tmp):
            for k, v in consts.items():
                if t.value == k:
                    tmp[i].type = v.type
                    tmp[i].value = v.value
        
        return tmp
